@@ -70,9 +70,30 @@ class OTReceiveConfigError(OTReceiveError):
 
 
 @dataclasses.dataclass
-class OTReceiveConfig:
+class OTReceiveRepoConfig:
     """OTReceiveRepo configuration
 
+    The path and url fields are required. See the OTReceiveConfig class for
+    details on the remaining optional fields.
+    """
+    path: Path
+    url: str
+    gpg_sign: list = dataclasses.field(default_factory=list)
+    gpg_homedir: str = None
+    gpg_verify: bool = False
+    gpg_trustedkeys: str = None
+    sign_type: str = 'ed25519'
+    sign_keyfiles: list = dataclasses.field(default_factory=list)
+    sign_verify: bool = False
+    sign_trustedkeyfile: str = None
+    update: bool = True
+    update_hook: str = None
+
+
+@dataclasses.dataclass
+class OTReceiveConfig:
+    """OTReceive configuration
+
     Configuration can be provided from a file or command line arguments using
     the load method. Config files are YAML mappings with the option names
     below using hypens instead of underscores. By default, the paths
@@ -107,6 +128,11 @@ class OTReceiveConfig:
       to the absolute path of the OSTree repository and the environment
       variable OSTREE_RECEIVE_REFS set to the set of refs received separated
       by whitespace.
+    repos: Optional per-repository configuration settings. All of the above
+      settings except for root can be set and will override the global value.
+      The value is a map of repository path to map of settings. The repository
+      path can be relative or absolute. If root is specified, relative paths
+      are resolved below it.
     log_level: Set the log level. See the logging module for available levels.
     force: Force receiving commits even if nothing changed or the remote
       commits are not newer than the current commits.
@@ -124,6 +150,7 @@ class OTReceiveConfig:
     sign_trustedkeyfile: str = None
     update: bool = True
     update_hook: str = None
+    repos: dict = dataclasses.field(default_factory=dict)
     log_level: str = 'INFO'
     force: bool = False
     dry_run: bool = False
@@ -217,8 +244,70 @@ class OTReceiveConfig:
             config_home / 'ostree/ostree-receive.conf',
         ]
 
+    def get_repo_config(self, path, url):
+        """Get OTReceiveRepoConfig instance for repo path and URL"""
+        repo_path = Path(path)
+        repo_root = (
+            Path(self.root).resolve() if self.root else None
+        )
+
+        if repo_root:
+            if not repo_path.is_absolute():
+                # Join the relative path to the root.
+                repo_path = repo_root.joinpath(repo_path)
+
+            # Make sure the path is below the root.
+            repo_path = repo_path.resolve()
+            try:
+                repo_path.relative_to(repo_root)
+            except ValueError:
+                raise OTReceiveError(f'repo {path} not found') from None
+
+        # Ensure the repository exists.
+        if not repo_path.exists():
+            raise OTReceiveError(f'repo {path} not found')
+
+        # See if there's a matching path in repos.
+        for key, values in self.repos.items():
+            config_path = Path(key)
+            if repo_root and not config_path.is_absolute():
+                config_path = repo_root.joinpath(config_path)
+            try:
+                matches = repo_path.samefile(config_path)
+            except FileNotFoundError:
+                matches = False
+
+            if matches:
+                logger.debug(f'Applying repos {key} configuration')
+                per_repo_config = values
+                break
+        else:
+            per_repo_config = {}
+
+        # Copy all but path and url from the per-repo or the global
+        # receive config.
+        repo_config_fields = {
+            field.name for field in dataclasses.fields(OTReceiveRepoConfig)
+        }
+        receive_config_fields = {
+            field.name for field in dataclasses.fields(self)
+        }
+        common_fields = repo_config_fields & receive_config_fields
+        repo_config_args = {
+            field: per_repo_config.get(field, getattr(self, field))
+            for field in common_fields
+        }
+        repo_config_args['path'] = repo_path
+        repo_config_args['url'] = url
+
+        return OTReceiveRepoConfig(**repo_config_args)
+
 
 class OTReceiveRepo(OSTree.Repo):
+    """OSTree repository receiving pushed commits
+
+    An OTReceiveRepoConfig instance is required.
+    """
     # The fake remote name
     REMOTE_NAME = '_receive'
 
@@ -229,32 +318,18 @@ class OTReceiveRepo(OSTree.Repo):
         OSTree.REPO_METADATA_REF,
     )
 
-    def __init__(self, path, url, config=None):
-        self.path = Path(path)
-        self.url = url
+    def __init__(self, config):
+        self.config = config
         self.remotes_dir = None
 
-        if config:
-            if not isinstance(config, OTReceiveConfig):
-                raise OTReceiveError(
-                    'config is not an OTReceiveConfig instance'
-                )
-            self.config = config
-        else:
-            self.config = OTReceiveConfig()
+        if not isinstance(self.config, OTReceiveRepoConfig):
+            raise OTReceiveError(
+                'config is not an OTReceiveRepoConfig instance'
+            )
 
-        if self.config.root:
-            repo_root = Path(self.config.root).resolve()
-            if not self.path.is_absolute():
-                # Join the relative path to the root.
-                self.path = repo_root.joinpath(self.path)
-
-            # Make sure the path is below the root.
-            self.path = self.path.resolve()
-            try:
-                self.path.relative_to(repo_root)
-            except ValueError:
-                raise OTReceiveError(f'repo {path} not found') from None
+        # Ensure the repository exists.
+        if not self.path.exists():
+            raise OTReceiveError(f'repo {self.path} not found')
 
         logger.debug('Using repo path %s', self.path)
 
@@ -292,6 +367,14 @@ class OTReceiveRepo(OSTree.Repo):
                          remotes_config_dir=self.remotes_dir.name)
         self.open()
 
+    @property
+    def path(self):
+        return self.config.path
+
+    @property
+    def url(self):
+        return self.config.url
+
     def __enter__(self):
         return self
 
@@ -533,10 +616,18 @@ class OTReceiveRepo(OSTree.Repo):
                     safe_sign_opts.append(f'--sign=<key #{i} from {keyfile}>')
 
         if self._is_flatpak_repo():
-            cmd_prefix = ['flatpak', 'build-update-repo', str(self.path)]
+            cmd_prefix = [
+                'flatpak',
+                'build-update-repo',
+                str(self.path),
+            ]
         else:
-            cmd_prefix = ['ostree', f'--repo={self.path}', 'summary',
-                          '--update']
+            cmd_prefix = [
+                'ostree',
+                f'--repo={self.path}',
+                'summary',
+                '--update',
+            ]
         logger.info('Updating repo metadata with %s',
                     ' '.join(cmd_prefix + safe_sign_opts))
         subprocess.check_call(cmd_prefix + sign_opts)
@@ -562,7 +653,7 @@ class OTReceiveRepo(OSTree.Repo):
         logger.debug('OSTREE_RECEIVE_REFS=%s', env['OSTREE_RECEIVE_REFS'])
         subprocess.check_call(cmd, env=env)
 
-    def receive(self, refs):
+    def receive(self, refs, force=False, dry_run=False):
         # See what revisions we're pulling.
         _, remote_refs = self.remote_list_refs(self.REMOTE_NAME)
         if len(refs) == 0:
@@ -596,7 +687,7 @@ class OTReceiveRepo(OSTree.Repo):
                 raise OTReceiveError(
                     f'Could not find ref {ref} in summary file')
 
-            if self.config.force or remote_rev != current_rev:
+            if force or remote_rev != current_rev:
                 logger.debug('Pulling %s', ref)
                 refs_to_pull[ref] = remote_rev
 
@@ -635,16 +726,17 @@ class OTReceiveRepo(OSTree.Repo):
                     else:
                         if remote_timestamp <= current_timestamp:
                             logger.warning(
-                                'ref %s remote rev %s is not newer than '
-                                'current rev %s',
-                                ref, remote_rev, current_rev
+                                'received %s commit %s is not newer than '
+                                'current %s commit %s',
+                                ref, remote_rev, ref, current_rev
                             )
                         if current_root.equal(remote_root):
                             logger.warning(
-                                'ref %s remote commit %s root equals %s',
-                                ref, remote_rev, current_rev
+                                'received %s commit %s has the same content '
+                                'as current %s commit %s',
+                                ref, remote_rev, ref, current_rev
                             )
-                        if self.config.force:
+                        if force:
                             logger.info('Forcing merge of ref %s', ref)
                             refs_to_merge[ref] = remote_rev
 
@@ -654,7 +746,7 @@ class OTReceiveRepo(OSTree.Repo):
                 return set()
 
             # For a dry run, exit now before creating the refs
-            if self.config.dry_run:
+            if dry_run:
                 self.abort_transaction()
                 return refs_to_merge.keys()
 
@@ -680,6 +772,30 @@ class OTReceiveRepo(OSTree.Repo):
         return refs_to_merge.keys()
 
 
+class OTReceiver:
+    """Pushed commit receiver
+
+    An OTReceiveConfig instance can be provided to configure the receiver.
+    """
+    def __init__(self, config=None):
+        self.config = config or OTReceiveConfig()
+
+        if not isinstance(self.config, OTReceiveConfig):
+            raise OTReceiveError(
+                'config is not an OTReceiveConfig instance'
+            )
+
+    def receive(self, path, url, refs):
+        """Receive pushed commits
+
+        Creates an OTReceiveRepo at path and receives commits on refs
+        from url.
+        """
+        repo_config = self.config.get_repo_config(path, url)
+        with OTReceiveRepo(repo_config) as repo:
+            return repo.receive(refs, self.config.force, self.config.dry_run)
+
+
 class OTReceiveArgParser(ArgumentParser):
     """ArgumentParse for ostree-receive"""
     def __init__(self):
@@ -725,8 +841,8 @@ def main():
 
     logging.basicConfig(level=config.log_level)
 
-    with OTReceiveRepo(args.repo, args.url, config) as repo:
-        repo.receive(args.refs)
+    receiver = OTReceiver(config)
+    receiver.receive(args.repo, args.url, args.refs)
 
 
 if __name__ == '__main__':
